"""
Sample code to render various representations.

Usage:
    python -m starter.render_generic --render point_cloud  # 5.1
    python -m starter.render_generic --render parametric  --num_samples 100  # 5.2
    python -m starter.render_generic --render implicit  # 5.3
"""
import argparse
import pickle

import matplotlib.pyplot as plt
import mcubes
import numpy as np
import pytorch3d
import torch

from starter.utils import get_device, get_mesh_renderer, get_points_renderer
from starter.utils import unproject_depth_image

import imageio
import math

def load_rgbd_data(path="data/rgbd_data.pkl"):
    with open(path, "rb") as f:
        data = pickle.load(f)
    return data

def get_cam(R0, T0, azim, device):
    R_rel = pytorch3d.transforms.axis_angle_to_matrix(torch.tensor([[0., azim * math.pi / 180., 0]])).squeeze(0)
    R = (R_rel @ R0).unsqueeze(0)
    # T = (R_rel @ T0).unsqueeze(0)
    T = T0.unsqueeze(0)
    cam = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, device=device)
    return cam

if __name__ == "__main__":
    data = load_rgbd_data()

    device = get_device()
    renderer = get_points_renderer(
        image_size=256, background_color=(1, 1, 1)
    )
    R, T = pytorch3d.renderer.look_at_view_transform(8, 10, 0)
    R_z = pytorch3d.transforms.axis_angle_to_matrix(torch.tensor([[0., 0, 180 * math.pi / 180.]])).squeeze(0)
    # R, T = R_z @ R[0], R_z @ T[0]
    R = R_z @ R[0]
    T = T[0]

    rgb1, mask1, depth1, cameras1 = data['rgb1'], data['mask1'], data['depth1'], data['cameras1']
    rgb2, mask2, depth2, cameras2 = data['rgb2'], data['mask2'], data['depth2'], data['cameras2']

    rgb1, mask1, depth1 = torch.tensor(rgb1), torch.tensor(mask1), torch.tensor(depth1)
    rgb2, mask2, depth2 = torch.tensor(rgb2), torch.tensor(mask2), torch.tensor(depth2)

    max_d = max(depth1.max(), depth2.max())
    # depth1 /= max_d
    # depth2 /= max_d

    points1, pts_rgb1 = unproject_depth_image(rgb1, mask1, depth1, cameras1)
    points2, pts_rgb2 = unproject_depth_image(rgb2, mask2, depth2, cameras2)
    points3, pts_rgb3 = torch.cat([points1, points2], dim=0), torch.cat([pts_rgb1, pts_rgb2], dim=0)

    pcd1 = pytorch3d.structures.Pointclouds(points=points1.unsqueeze(0), features=pts_rgb1.unsqueeze(0)).to(device)
    pcd2 = pytorch3d.structures.Pointclouds(points=points2.unsqueeze(0), features=pts_rgb2.unsqueeze(0)).to(device)
    pcd3 = pytorch3d.structures.Pointclouds(points=points3.unsqueeze(0), features=pts_rgb3.unsqueeze(0)).to(device)

    azims = torch.linspace(0, 360, 6)
    # azims = [60]

    rend_list = []
    for azim in azims:

        cam = get_cam(R, T, azim, device)
        rend1 = renderer(pcd1, cameras=cam).cpu().numpy()[0, ..., :3]
        rend2 = renderer(pcd2, cameras=cam).cpu().numpy()[0, ..., :3]
        rend3 = renderer(pcd3, cameras=cam).cpu().numpy()[0, ..., :3]

        rend_all = np.concatenate([rend1, rend2, rend3], axis=1)
        rend_all = (rend_all * 255).astype(np.uint8)
        rend_list.append(rend_all)

        # out_path = f'output/q5.1-{azim}.png'
        # plt.imsave(out_path, rend_all)
    imageio.mimsave(f'output/q5.1-all.gif', rend_list, fps=2)

