"""
Sample code to render various representations.

Usage:
    python -m starter.render_generic --render point_cloud  # 5.1
    python -m starter.render_generic --render parametric  --num_samples 100  # 5.2
    python -m starter.render_generic --render implicit  # 5.3
"""
import argparse
import pickle

import matplotlib.pyplot as plt
import mcubes
import numpy as np
import pytorch3d
import torch

from starter.utils import get_device, get_mesh_renderer, get_points_renderer
from starter.utils import unproject_depth_image

from utils_yc import get_cam

import imageio
import math

if __name__ == "__main__":
    device = get_device()
    image_size = 256
    renderer = get_mesh_renderer(image_size=image_size, device=device)

    dist = 3
    elev = 0
    R, T = pytorch3d.renderer.look_at_view_transform(dist=dist, elev=elev, azim=0)
    cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, device=device)

    R_torus = 0.6
    r_torus = R_torus / 3 * 2
    # r_torus = 0.2

    min_value = -1.1
    max_value = 1.1
    voxel_size = 64
    X, Y, Z = torch.meshgrid([torch.linspace(min_value, max_value, voxel_size)] * 3)
    # voxels = (torch.sqrt(X ** 2 + Z ** 2) - R_torus) ** 2 + Y ** 2 - r_torus ** 2
    voxels = (torch.sqrt(X ** 2 + Y ** 2) - R_torus) ** 2 + Z ** 2 - r_torus ** 2

    vertices, faces = mcubes.marching_cubes(mcubes.smooth(voxels), isovalue=0)
    vertices = torch.tensor(vertices).float()
    faces = torch.tensor(faces.astype(int))

    vertices = (vertices / voxel_size) * (max_value - min_value) + min_value
    # textures = (vertices - vertices.min()) / (vertices.max() - vertices.min())
    # textures = pytorch3d.renderer.TexturesVertex(vertices.unsqueeze(0))

    vertices = vertices.unsqueeze(0)
    faces = faces.unsqueeze(0)

    N = vertices.shape[1]
    color1 = torch.tensor([1.0, 0.0, 0.0]).repeat(N, 1)
    color2 = torch.tensor([0.0, 0.0, 1.0]).repeat(N, 1)

    z_min = vertices[0, :, 2].min()
    z_max = vertices[0, :, 2].max()
    alpha = (vertices[0, :, 2] - z_min) / (z_max - z_min)
    alpha = alpha.unsqueeze(1)
    color = alpha * color2 + (1-alpha) * color1
    textures = pytorch3d.renderer.TexturesVertex(color.unsqueeze(0))

    mesh = pytorch3d.structures.Meshes(vertices, faces, textures=textures).to(
        device
    )
    # lights = pytorch3d.renderer.PointLights(location=[[0, 0.0, -dist]], device=device,)
    lights = pytorch3d.renderer.PointLights(location=-T, device=device,)

    azims = torch.linspace(0, 360, 36)

    rend_list = []

    for azim in azims:

        # look_at_view
        R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
        cam = cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)

        rend = renderer(mesh, cameras=cameras, lights=lights).cpu().numpy()[0, ..., :3]
        rend = (rend * 255).astype(np.uint8)
        rend_list.append(rend)

    imageio.mimsave(f'output/q5.3-torus-implicit.gif', rend_list, fps=15)