import os
import numpy as np
import imageio

import math

import torch
import pytorch3d

from starter.utils import get_device, get_mesh_renderer, load_cow_mesh, get_points_renderer

from utils_yc import render_360

def mesh_to_pcd(mesh, N_pts=100):
    verts = mesh.verts_list()[0]
    faces = mesh.faces_list()[0]
    areas = mesh.faces_areas_packed()
    
    # number to sample
    prob = areas / areas.sum()
    sampled_faces_ix = prob.multinomial(num_samples=N_pts, replacement=True)
    sampled_faces = faces[sampled_faces_ix]
    
    sampled_verts = verts[sampled_faces] # N_f, 3 (v1, v2, v3), 3 (x, y, z)
    
    alpha = torch.rand(N_pts)
    beta = (1-alpha) * torch.rand(N_pts)
    gamma = 1 - alpha - beta
    # print(alpha + beta + gamma
    abg = torch.stack([alpha, beta, gamma], dim=-1).unsqueeze(-1).to(device)

    verts_new = (sampled_verts * abg).sum(1)
    feats_new = torch.ones_like(verts_new)
    pcd = pytorch3d.structures.Pointclouds(points=verts_new.unsqueeze(0), features=feats_new.unsqueeze(0)).to(device)
    return pcd

# def make_batched(struct, b=10):
#     device = struct.device
#     if isinstance(struct, pytorch3d.structures.meshes.Meshes):
#         mesh = struct
#         verts = mesh.verts_list()[0].repeat(b, 1, 1).cpu()
#         faces = mesh.faces_list()[0].repeat(b, 1, 1).cpu()
#         textures = mesh.textures.verts_features_list()[0].repeat(b, 1, 1).cpu()  # (1, N_v, 3)
#         import pdb; pdb.set_trace()
#         ret = pytorch3d.structures.Meshes(
#             verts=vertices,
#             faces=faces,
#             textures=pytorch3d.renderer.TexturesVertex(textures),
#         ).to(device)
#     else:
#         pcd = struct
#         verts = pcd.points_list()[0].repeat(b, 1, 1)
#         color = (verts - verts.min()) / (verts.max() - verts.min())
#         ret = pytorch3d.structures.Pointclouds(points=verts, features=color).to(device)
        
#     return ret
    

if __name__ == '__main__':

    out_d = 'output'
    if not os.path.exists(out_d): os.makedirs(out_d)

    device = get_device()
    cow_path = 'data/cow.obj'

    # print(device)

    # renderer
    image_size = 256
    renderer = get_points_renderer(
        image_size=256, background_color=(1, 1, 1)
    )
    
    mesh_renderer = get_mesh_renderer(image_size=image_size)

    # Get the vertices, faces, and textures.
    vertices, faces = load_cow_mesh(cow_path)
    vertices = vertices.unsqueeze(0)  # (N_v, 3) -> (1, N_v, 3)
    faces = faces.unsqueeze(0)  # (N_f, 3) -> (1, N_f, 3)
    textures = torch.ones_like(vertices)  # (1, N_v, 3)
    color=[0.7, 0.7, 1]
    textures = textures * torch.tensor(color)  # (1, N_v, 3)
    mesh = pytorch3d.structures.Meshes(
        verts=vertices,
        faces=faces,
        textures=pytorch3d.renderer.TexturesVertex(textures),
    )
    mesh = mesh.to(device)
    
    N_pts_list = [10, 100, 1000, 10000]
    # pcd_list = [mesh_to_pcd(mesh, N) for N in N_pts_list]
    pcd_list = []
    
    for n in N_pts_list:
        pcd_list.append(mesh_to_pcd(mesh, n))
    
    shapes = [mesh] + pcd_list
    
    N_views = 36
    dist = 3
    elev = 10
    azims = torch.linspace(-180, 180, N_views)
    
    R, T = pytorch3d.renderer.look_at_view_transform(dist, elev, azims)
    cams = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, device=device)
    
    imgs = []
    for i in range(len(shapes)):
        s = shapes[i].clone()
        # s = make_batched(s, N_views)
        s_batched = s.extend(N_views)
            
        if i == 0:
            rend = mesh_renderer(s_batched, cameras=cams)
        else:
            rend = renderer(s_batched, cameras=cams)
        
        rend = rend.cpu().numpy()[..., :3]
        rend = (rend * 255).astype(np.uint8)
        
        imgs.append(rend)
    
    imgs = np.concatenate(imgs, axis=2)
    imgs = [im for im in imgs]
    # pcd_all = make_batched(pcd_list[-1], N_views)
    # rend1 = renderer(pcd_all, cameras=cams).cpu().numpy()#[0, ..., :3]
    # rend1 = (rend1 * 255).astype(np.uint8)
    # all_views = [r for r in rend1]
    
    imageio.mimsave(f'output/q7.gif', imgs, fps=15)
    
    # imageio.mimsave(f'{out_d}/q1.1_cow_gif.gif', rend1, fps=15)